"""
This module implements some common finite difference schemes
"""
import numpy as np
from pynumdiff.utils import utility


def first_order(x, dt, params=None, options=None):
    """
    :param x: (np.array of floats, 1xN) time series to differentiate
    :param dt: (float) time step
    :param params: (list)  [iterations] : (int, optional)  number of iterations
                                            (ignored if 'iterate' not in options)
    :param options: (dict)  {'iterate'}  : (bool, optional) iterate the finite difference
                                            method (smooths the estimates)
    :return:    x_hat    : estimated (smoothed) x
                dxdt_hat : estimated derivative of x
    """

    if options and 'iterate' in options:
        assert params and isinstance(params, list), "params should be a non-empty list!"
        return __iterate_first_order__(x, dt, params)

    # Calculate the finite difference
    dxdt_hat = np.diff(x) / dt
    # Pad the data
    dxdt_hat = np.hstack((dxdt_hat[0], dxdt_hat, dxdt_hat[-1]))
    # Re-finite dxdt_hat using linear interpolation
    dxdt_hat = np.mean((dxdt_hat[0:-1], dxdt_hat[1:]), axis=0)

    return x, dxdt_hat


def second_order(x, dt):
    """
    :param x: (np.array of floats, 1xN) time series to differentiate
    :param dt: (float) time step
    :return:    x_hat: estimated (smoothed) x
                dxdt_hat: estimated derivative of x
    """
    dxdt_hat = (x[2:] - x[0:-2]) / (2 * dt)
    first_dxdt_hat = (-3 * x[0] + 4 * x[1] - x[2]) / (2 * dt)
    last_dxdt_hat = (3 * x[-1] - 4 * x[-2] + x[-3]) / (2 * dt)
    dxdt_hat = np.hstack((first_dxdt_hat, dxdt_hat, last_dxdt_hat))
    return x, dxdt_hat


def __x_hat_using_finite_difference__(x, dt):
    """
    :param x:
    :param dt:
    :return:
    """
    x_hat, dxdt_hat = first_order(x, dt)
    x_hat = utility.integrate_dxdt_hat(dxdt_hat, dt)
    x0 = utility.estimate_initial_condition(x, x_hat)
    return x_hat + x0


def __iterate_first_order__(x, dt, params):
    """
    Iterate first order centered finite difference.
    :param x: (np.array of floats, 1xN) time series to differentiate
    :param dt: (float) time step
    :param params: (list)  [iterations] : (int) number of iterations
    :return:    x_hat: estimated (smoothed) x
                dxdt_hat: estimated derivative of x
    """
    if isinstance(params, list):
        iterations = params[0]
    else:
        iterations = params

    # set up weights
    w = np.arange(0, len(x), 1)
    w = w / np.max(w)

    # forward backward passes
    for _ in range(iterations):
        xf = __x_hat_using_finite_difference__(x, dt)
        xb = __x_hat_using_finite_difference__(x[::-1], dt)
        x = xf * w + xb[::-1] * (1 - w)

    x_hat, dxdt_hat = first_order(x, dt)

    return x_hat, dxdt_hat



from __future__ import absolute_import, division, print_function
from os.path import join as pjoin

# Format expected by setup.py and doc/source/conf.py: string of form "X.Y.Z"
_version_major = 0
_version_minor = 1
_version_micro = ''  # use '' for first of series, number for 1 and above
_version_extra = 'dev'

# Construct full version string from these.
_ver = [_version_major, _version_minor]
if _version_micro:
    _ver.append(_version_micro)
if _version_extra:
    _ver.append(_version_extra)

__version__ = '.'.join(map(str, _ver))

CLASSIFIERS = ["Development Status :: 3 - Alpha",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: MIT License",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering"]

# Description should be a one-liner:
description = "pynumdiff: numerical derivatives in python"
# Long description will go up on the pypi page
long_description = """
PyNumDiff
=========
PyNumDiff (pynumdiff) is a package for calculating numerical derivatives
in python, using a variety of different methods.
"""

NAME = "pynumdiff"
MAINTAINER = "Yuying Liu"
MAINTAINER_EMAIL = "yliu814@uw.edu"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = "https://github.com/luckystarufo/PyNumDiff/tree/upgrade"
DOWNLOAD_URL = ""
LICENSE = "MIT"
AUTHOR = "Floris van Breugel, Yuying Liu"
AUTHOR_EMAIL = "florisvb@gmail.com, liuyuyingufo@gmail.com"
PLATFORMS = "OS Independent"
MAJOR = _version_major
MINOR = _version_minor
MICRO = _version_micro
VERSION = __version__
PACKAGE_DATA = {'pynumdiff': [pjoin('data', '*')]}
REQUIRES = ["numpy", "scipy", "matplotlib"]
OPTIONAL_REQUIREMENTS = ["pydmd", "cvxpy", "pychebfun"]


"""
All kinds of utilities
"""
import os
import sys
import copy
import numpy as np
import scipy


def get_filenames(path, contains, does_not_contain=('~', '.pyc')):
    """
    :param path:
    :param contains:
    :param does_not_contain:
    :return:
    """
    cmd = 'ls ' + '"' + path + '"'
    ls = os.popen(cmd).read()
    all_filelist = ls.split('\n')
    all_filelist.remove('')
    filelist = []
    for _, filename in enumerate(all_filelist):
        if contains in filename:
            fileok = True
            for nc in does_not_contain:
                if nc in filename:
                    fileok = False
            if fileok:
                filelist.append( os.path.join(path, filename) )
    return filelist


def is_odd(num):
    """
    :param num:
    :return:
    """
    return num & 0x1


def isnotebook():
    """
    Checks to see if the environment is an interactive notebook or not.
    :return: True or False
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        # elif shell == 'TerminalInteractiveShell':
        #     return False  # Terminal running IPython
        return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


def hankel_matrix(x, num_delays, pad=False):  # fixed delay step of 1
    """
    :param x: numpy array or matrix
    :param num_delays: int, number of times to 1-step shift data
    :param pad:
    :return: a Hankel Matrix m

            e.g.  if
                    x = [a, b, c, d, e] and num_delays = 3
            then with pad = False:
                    m = [['a', 'b', 'c'],
                         ['b', 'c', 'd'],
                         ['c', 'd', 'e']]
            or pad = True:
                    m = [['a', 'b', 'c', 'd', 'e'],
                         ['b', 'c', 'd', 'e',  0],
                         ['c', 'd', 'e',  0,   0]]
    """

    m = copy.copy(x)
    for d in range(1, num_delays):
        xi = x[:, d:]
        xi = np.pad(xi, ((0, 0), (0, x.shape[1]-xi.shape[1])), 'constant', constant_values=0)
        m = np.vstack((m, xi))
    if not pad:
        return m[:, 0:-1*num_delays]
    return m


def matrix_inv(X, max_sigma=1e-16):
    """
    :param X:
    :param max_sigma:
    :return:
    """
    U, Sigma, V = np.linalg.svd(X, full_matrices=False)
    Sigma_inv = Sigma**-1
    Sigma_inv[np.where(Sigma < max_sigma)[0]] = 0  # helps reduce instabilities
    return V.T.dot(np.diag(Sigma_inv)).dot(U.T)


def total_variation(x):
    """
    :param x:
    :return:
    """
    if np.isnan(x).any():
        return np.nan
    x1 = np.ravel(x)[0:-1]
    x2 = np.ravel(x)[1:]
    return np.sum(np.abs(x2-x1))/len(x1)  # mostly equivalent to cvxpy.tv(x2-x1).value


def peakdet(v, delta, x=None):
    """
    Converted from MATLAB script at http://billauer.co.il/peakdet.html
    Returns two arrays
    function [maxtab, mintab]=peakdet(v, delta, x)
    %PEAKDET Detect peaks in a vector
    %        [MAXTAB, MINTAB] = PEAKDET(V, DELTA) finds the local
    %        maxima and minima ("peaks") in the vector V.
    %        MAXTAB and MINTAB consists of two columns. Column 1
    %        contains indices in V, and column 2 the found values.
    %        With [MAXTAB, MINTAB] = PEAKDET(V, DELTA, X) the indices
    %        in MAXTAB and MINTAB are replaced with the corresponding
    %        X-values.
    %
    %        A point is considered a maximum peak if it has the maximal
    %        value, and was preceded (to the left) by a value lower by
    %        DELTA.
    % Eli Billauer, 3.4.05 (Explicitly not copyrighted).
    % This function is released to the public domain; Any use is allowed.

    :param v:
    :param delta:
    :param x:
    :return:

    """
    maxtab = []
    mintab = []
    if x is None:
        x = np.arange(len(v))
    v = np.asarray(v)
    if len(v) != len(x):
        sys.exit('Input vectors v and x must have same length')
    if not np.isscalar(delta):
        sys.exit('Input argument delta must be a scalar')
    if delta <= 0:
        sys.exit('Input argument delta must be positive')

    mn, mx = np.Inf, -1*np.Inf
    mnpos, mxpos = np.NaN, np.NaN
    lookformax = True
    for i in np.arange(len(v)):
        this = v[i]
        if this > mx:
            mx = this
            mxpos = x[i]
        if this < mn:
            mn = this
            mnpos = x[i]
        if lookformax:
            if this < mx-delta:
                maxtab.append((mxpos, mx))
                mn = this
                mnpos = x[i]
                lookformax = False
        else:
            if this > mn+delta:
                mintab.append((mnpos, mn))
                mx = this
                mxpos = x[i]
                lookformax = True

    return np.array(maxtab), np.array(mintab)


# simple finite difference
def finite_difference(x, dt):
    """
    :param x: (np.array of floats, 1xN) time series to differentiate
    :param dt: (float) time step
    :param params: (list): (int, optional) number of iterations ignored if 'iterate' not in options
    :param options: (bool) iterate the finite difference method (smooths the estimates)
    :return: x_hat: smoothed x; dxdt_hat: derivative of x
    """
    # Calculate the finite difference
    dxdt_hat = np.diff(x)/dt
    # Pad the data
    dxdt_hat = np.hstack((dxdt_hat[0], dxdt_hat, dxdt_hat[-1]))
    # Re-finite dxdt_hat using linear interpolation
    dxdt_hat = np.mean((dxdt_hat[0:-1], dxdt_hat[1:]), axis=0)

    return x, dxdt_hat


# Trapazoidal integration, with interpolated final point so that the lengths match.
def integrate_dxdt_hat(dxdt_hat, dt):
    """
    :param dxdt_hat:
    :param dt:
    :return:
    """
    x = scipy.integrate.cumtrapz(dxdt_hat)
    first_value = x[0] - np.mean(dxdt_hat[0:1])
    x = np.hstack((first_value, x))*dt
    return x


# Optimization routine to estimate the integration constant.
def estimate_initial_condition(x, x_hat):
    """
    :param x:
    :param x_hat:
    :return:
    """
    def f(x0, *args):
        x, x_hat = args[0]
        error = np.linalg.norm(x - (x_hat+x0))
        return error
    result = scipy.optimize.minimize(f, [0], args=[x, x_hat], method='SLSQP')
    return result.x


# kernels
def __mean_kernel__(window_size):
    """
    :param window_size:
    :return:
    """
    return np.ones(window_size)/window_size


def __gaussian_kernel__(window_size):
    """
    :param window_size:
    :return:
    """
    sigma = window_size / 6.
    t = np.linspace(-2.7*sigma, 2.7*sigma, window_size)
    gaussian_func = lambda t, sigma: 1/np.sqrt(2*np.pi*sigma**2) * np.exp(-(t**2)/(2*sigma**2))
    ker = gaussian_func(t, sigma)
    return ker / np.sum(ker)


def __friedrichs_kernel__(window_size):
    """
    :param window_size:
    :return:
    """
    x = np.linspace(-0.999, 0.999, window_size)
    ker = np.exp(-1/(1-x**2))
    return ker / np.sum(ker)



"""
Simulation Related
"""
import numpy as _np
from scipy.integrate import odeint

# local imports
from pynumdiff.utils import utility as _utility
from pynumdiff.utils import _pi_cruise_control
_finite_difference = _utility.finite_difference


# pylint: disable-msg=too-many-locals, too-many-arguments
def __add_noise__(x, noise_type, noise_parameters, random_seed):
    """
    :param x:
    :param noise_type:
    :param noise_parameters:
    :param random_seed:
    :return:
    """
    _np.random.seed(random_seed)
    timeseries_length = _np.max(x.shape)
    noise = _np.random.__getattribute__(noise_type)(noise_parameters[0], noise_parameters[1],
                                                    timeseries_length)
    return x + noise


def sine(timeseries_length=4, noise_type='normal', noise_parameters=(0, 0.5), random_seed=1,
         dt=0.01, simdt=0.0001, frequencies=(1, 1.7), magnitude=1):
    """
    :param timeseries_length:
    :param noise_type:
    :param noise_parameters:
    :param random_seed:
    :param dt:
    :param simdt:
    :param frequencies:
    :param magnitude:
    :return:
    """
    y_offset = 1

    t = _np.arange(0, timeseries_length, simdt)
    x = y_offset
    dxdt = 0
    for f in frequencies:
        x += magnitude/len(frequencies)*_np.sin(t*2*_np.pi*f)
        dxdt += magnitude/len(frequencies)*_np.cos(t*2*_np.pi*f)*2*_np.pi*f
    actual_vals = _np.matrix(_np.vstack((x, dxdt)))
    noisy_x = __add_noise__(x, noise_type, noise_parameters, random_seed)
    #
    noisy_measurements = _np.matrix(noisy_x)
    #
    noisy_pos = _np.ravel(noisy_measurements)
    pos = _np.ravel(actual_vals[0, :])
    vel = _np.ravel(actual_vals[1, :])

    idx = _np.arange(0, len(t), int(dt/simdt))
    return noisy_pos[idx], pos[idx], vel[idx], None


def large_sine(timeseries_length=4, noise_type='normal', noise_parameters=(0, 0.5), random_seed=1,
               dt=0.01, simdt=0.0001, frequencies=(7, 10), magnitude=100):
    """
    :param timeseries_length:
    :param noise_type:
    :param noise_parameters:
    :param random_seed:
    :param dt:
    :param simdt:
    :param frequencies:
    :param magnitude:
    :return:
    """
    y_offset = 1

    t = _np.arange(0, timeseries_length, simdt)
    x = y_offset
    dxdt = 0
    for f in frequencies:
        x += magnitude/len(frequencies)*_np.sin(t*2*_np.pi*f)
        dxdt += magnitude/len(frequencies)*_np.cos(t*2*_np.pi*f)*2*_np.pi*f
    actual_vals = _np.matrix(_np.vstack((x, dxdt)))

    noisy_x = __add_noise__(x, noise_type, noise_parameters, random_seed)

    noisy_measurements = _np.matrix(noisy_x)

    noisy_pos = _np.ravel(noisy_measurements)
    pos = _np.ravel(actual_vals[0,:])
    vel = _np.ravel(actual_vals[1,:])

    idx = _np.arange(0, len(t), int(dt/simdt))
    return noisy_pos[idx], pos[idx], vel[idx], None


def triangle(timeseries_length=4, noise_type='normal', noise_parameters=(0, 0.5), random_seed=1,
             dt=0.01, simdt=0.0001):
    """
    :param timeseries_length:
    :param noise_type:
    :param noise_parameters:
    :param random_seed:
    :param dt:
    :param simdt:
    :return:
    """
    t = _np.arange(0, timeseries_length, simdt)
    continuous_x = _np.sin(t*t)

    # find peaks and valleys
    peaks, valleys = _utility.peakdet(continuous_x, 0.1)

    # organize peaks and valleys
    if len(peaks) > 0:
        reversal_idxs = peaks[:,0].astype(int).tolist()
        reversal_idxs.extend(valleys[:, 0].astype(int).tolist())
        reversal_idxs.extend([0, len(continuous_x)-1])

        reversal_vals = peaks[:,1].tolist()
        reversal_vals.extend(valleys[:, 1].tolist())
        reversal_vals.extend([0, continuous_x[-1]])
    else:
        reversal_idxs = [0, len(continuous_x)-1]
        reversal_vals = [0, continuous_x[-1]]

    idx = _np.argsort(reversal_idxs)
    reversal_idxs = _np.array(reversal_idxs)[idx]
    reversal_vals = _np.array(reversal_vals)[idx]
    reversal_ts = t[reversal_idxs]

    x = _np.interp(t, reversal_ts, reversal_vals)
    _, dxdt = _finite_difference(x, dt=simdt)

    noisy_x = __add_noise__(x, noise_type, noise_parameters, random_seed)

    actual_vals = _np.matrix(_np.vstack((x, dxdt)))
    noisy_measurements = _np.matrix(noisy_x)

    noisy_pos = _np.ravel(noisy_measurements)
    pos = _np.ravel(actual_vals[0,:])
    vel = _np.ravel(actual_vals[1,:])

    idx = _np.arange(0, len(t), int(dt/simdt))
    return noisy_pos[idx], pos[idx], vel[idx], None


def pop_dyn(timeseries_length=4, noise_type='normal', noise_parameters=(0, 0.5), random_seed=1,
            dt=0.01, simdt=0.0001):
    """
    :param timeseries_length:
    :param noise_type:
    :param noise_parameters:
    :param random_seed:
    :param dt:
    :param simdt:
    :return:
    """
    # http://www.biologydiscussion.com/population/population-growth/population-growth-curves-ecology/51854
    t = _np.arange(0, timeseries_length, simdt)
    K = 2  # carrying capacity
    r = 4  # biotic potential

    x = [0.1] # population
    dxdt = [r*x[-1]*(1-x[-1]/K)]
    for _ in t[1:]:
        x.append(x[-1] + simdt*dxdt[-1])
        dxdt.append(r*x[-1]*(1-x[-1]/K))

    x = _np.matrix(x)
    dxdt = _np.matrix(dxdt)

    noisy_x = __add_noise__(x, noise_type, noise_parameters, random_seed)

    actual_vals = _np.matrix(_np.vstack((x, dxdt)))
    noisy_measurements = _np.matrix(noisy_x)

    noisy_pos = _np.ravel(noisy_measurements)
    pos = _np.ravel(actual_vals[0,:])
    vel = _np.ravel(actual_vals[1,:])

    idx = _np.arange(0, len(t), int(dt/simdt))
    return noisy_pos[idx], pos[idx], vel[idx], None


def linear_autonomous(timeseries_length=4, noise_type='normal', noise_parameters=(0, 0.5),
                      random_seed=1, dt=0.01, simdt=0.0001):
    """
    :param timeseries_length:
    :param noise_type:
    :param noise_parameters:
    :param random_seed:
    :param dt:
    :param simdt:
    :return:
    """
    t = _np.arange(0, timeseries_length, simdt)

    A = _np.matrix([[1, simdt, 0], [0, 1, simdt], [-100, -3, 0.01]])
    print(A)
    x0 = _np.matrix([[0], [2], [0]])
    xs = x0
    for _ in t:
        x = A*xs[:,-1]
        xs = _np.hstack((xs, x))

    x = xs[0,:]
    x *= 2

    smooth_x, dxdt = _finite_difference( _np.ravel(x), simdt)
    noisy_x = __add_noise__(x, noise_type, noise_parameters, random_seed)

    idx = _np.arange(0, len(t), int(dt/simdt))
    return _np.ravel(noisy_x)[1:][idx], smooth_x[1:][idx], dxdt[1:][idx], None


def pi_control(timeseries_length=4, noise_type='normal', noise_parameters=(0, 0.5),
               random_seed=1, dt=0.01, simdt=0.0001):
    """
    :param timeseries_length:
    :param noise_type:
    :param noise_parameters:
    :param random_seed:
    :param dt:
    :param simdt:
    :return:
    """
    t = _np.arange(0, timeseries_length, simdt)

    actual_vals, extra_measurements, controls = _pi_cruise_control.run(timeseries_length, simdt)
    x = actual_vals[0,:]
    dxdt = actual_vals[1,:]

    noisy_x = __add_noise__(x, noise_type, noise_parameters, random_seed)

    actual_vals = _np.matrix(_np.vstack((x, dxdt)))
    noisy_measurements = _np.matrix(noisy_x)

    noisy_pos = _np.ravel(noisy_measurements)
    pos = _np.ravel(actual_vals[0,:])
    vel = _np.ravel(actual_vals[1,:])

    idx = _np.arange(0, len(t), int(dt/simdt))
    return noisy_pos[idx], pos[idx], vel[idx], \
           [_np.array(extra_measurements)[0, idx], _np.array(controls)[0, idx]]


def lorenz_x(timeseries_length=4, noise_type='normal', noise_parameters=(0, 0.5),
             random_seed=1, dt=0.01, simdt=0.0001):
    """
    :param timeseries_length:
    :param noise_type:
    :param noise_parameters:
    :param random_seed:
    :param dt:
    :param simdt:
    :return:
    """
    noisy_measurements, actual_vals, _ = lorenz_xyz(timeseries_length, noise_type, noise_parameters,
                                                    random_seed, dt, simdt)

    noisy_pos = _np.ravel(noisy_measurements[0, :])
    pos = _np.ravel(actual_vals[0,:])
    vel = _np.ravel(actual_vals[3,:])

    return noisy_pos, pos, vel, None


def lorenz_xyz(timeseries_length=4, noise_type='normal', noise_parameters=(0, 0.5), random_seed=1,
               dt=0.01, simdt=0.0001, x0=(5, 1, 3), normalize=True):
    """
    :param timeseries_length:
    :param noise_type:
    :param noise_parameters:
    :param random_seed:
    :param dt:
    :param simdt:
    :param x0:
    :param normalize:
    :return:
    """
    t = _np.arange(0, timeseries_length, simdt)

    sigma = 10
    beta = 8/3
    rho = 45

    x = x0[0]
    y = x0[1]
    z = x0[2]
    xyz = _np.matrix([[x], [y], [z]])
    xyz_dot = None

    for _ in t:
        x, y, z = _np.ravel(xyz[:, -1])

        xdot = sigma*(y-x)
        ydot = x*(rho-z)-y
        zdot = x*y - beta*z

        new_xyz_dot = _np.matrix([[xdot], [ydot], [zdot]])
        if xyz_dot is None:
            xyz_dot = new_xyz_dot
        else:
            xyz_dot = _np.hstack((xyz_dot, new_xyz_dot))

        new_xyz = xyz[:,-1] + simdt*new_xyz_dot
        xyz = _np.hstack((xyz, new_xyz))

    if normalize:
        f = 20
    else:
        f = 1

    x = xyz[0,0:-1] / f
    dxdt = xyz_dot[0,:] / f

    y = xyz[1,0:-1] / f
    dydt = xyz_dot[1,:] / f

    z = xyz[2,0:-1] / f
    dzdt = xyz_dot[2,:] / f

    noisy_x = __add_noise__(x, noise_type, noise_parameters, random_seed)
    noisy_y = __add_noise__(y, noise_type, noise_parameters, random_seed+1)
    noisy_z = __add_noise__(z, noise_type, noise_parameters, random_seed+2)

    actual_vals = _np.array(_np.vstack((x, y, z, dxdt, dydt, dzdt)))
    noisy_measurements = _np.array(_np.vstack((noisy_x, noisy_y, noisy_z)))

    idx = _np.arange(0, len(t), int(dt/simdt))
    return noisy_measurements[:, idx], actual_vals[:, idx], None


def rk4_lorenz_xyz(timeseries_length=4, noise_type='normal', noise_parameters=(0, 0.5),
                   random_seed=1, dt=0.01, normalize=True):
    """
    :param timeseries_length:
    :param noise_type:
    :param noise_parameters:
    :param random_seed:
    :param dt:
    :param normalize:
    :return:
    """

    sigma = 10
    beta = 8/3
    rho = 45

    def dxyz_dt(xyz):
        x, y, z = xyz
        xdot = sigma*(y-x)
        ydot = x*(rho-z)-y
        zdot = x*y - beta*z
        return [xdot, ydot, zdot]

    ts = _np.linspace(0, timeseries_length, timeseries_length/dt)
    xyz_0 = [5, 1, 3]

    vals, _ = odeint(dxyz_dt, xyz_0, ts, full_output=True)
    vals = vals.T

    if normalize:
        x = vals[0, :]/20.
        y = vals[1, :]/20.
        z = vals[2, :]/20.
    else:
        x = vals[0, :]
        y = vals[1, :]
        z = vals[2, :]

    noisy_x = __add_noise__(x, noise_type, noise_parameters, random_seed)
    noisy_y = __add_noise__(y, noise_type, noise_parameters, random_seed+1)
    noisy_z = __add_noise__(z, noise_type, noise_parameters, random_seed+2)

    _, dxdt = _finite_difference(x, dt)
    _, dydt = _finite_difference(y, dt)
    _, dzdt = _finite_difference(z, dt)

    actual_vals = _np.array(_np.vstack((x, y, z, dxdt, dydt, dzdt)))
    noisy_measurements = _np.array(_np.vstack((noisy_x, noisy_y, noisy_z)))

    return noisy_measurements, actual_vals, None



"""
Metrics and evaluations?
"""
import numpy as _np
import matplotlib.pyplot as _plt
import scipy.stats as _scipy_stats

# local imports
from pynumdiff.utils import utility as _utility
_finite_difference = _utility.finite_difference


# pylint: disable-msg=too-many-locals, too-many-arguments
def plot(x, dt, x_hat, dxdt_hat, x_truth, dxdt_truth, xlim=None, ax_x=None, ax_dxdt=None,
         show_error=True, markersize=5):
    """
    :param x:
    :param dt:
    :param x_hat:
    :param dxdt_hat:
    :param x_truth:
    :param dxdt_truth:
    :param xlim:
    :param ax_x:
    :param ax_dxdt:
    :param show_error:
    :param markersize:
    :return:
    """
    t = _np.arange(0, dt*len(x), dt)
    if ax_x is None and ax_dxdt is None:
        fig = _plt.figure(figsize=(20,6))
        ax_x = fig.add_subplot(121)
        ax_dxdt = fig.add_subplot(122)

    if xlim is None:
        xlim = [t[0], t[-1]]

    if ax_x is not None:
        if x_hat is not None:
            ax_x.plot(t, x_hat, color='red')
        ax_x.plot(t, x_truth, '--', color='black')
        ax_x.plot(t, x, '.', color='blue', zorder=-100, markersize=markersize)
        ax_x.set_ylabel('Position')
        ax_x.set_xlabel('Time')
        ax_x.set_xlim(xlim[0], xlim[-1])

    if ax_dxdt is not None:
        ax_dxdt.plot(t, dxdt_hat, color='red')
        ax_dxdt.plot(t, dxdt_truth, '--', color='black')
        ax_dxdt.set_ylabel('Velocity')
        ax_dxdt.set_xlabel('Time')
        ax_dxdt.set_xlim(xlim[0], xlim[-1])

    if show_error:
        _, _, rms_dxdt = metrics(x, dt, x_hat, dxdt_hat, x_truth, dxdt_truth)
        print('RMS error in velocity: ', rms_dxdt)


def __rms_error__(a, e):
    """
    :param a:
    :param e:
    :return:
    """
    if _np.max(_np.abs(a-e)) > 1e16:
        return 1e16
    s_error = _np.ravel((a - e))**2
    ms_error = _np.mean(s_error)
    rms_error = _np.sqrt(ms_error)
    return rms_error


def metrics(x, dt, x_hat, dxdt_hat, x_truth=None, dxdt_truth=None, padding=None):
    """
    :param x:
    :param dt:
    :param x_hat:
    :param dxdt_hat:
    :param x_truth:
    :param dxdt_truth:
    :param padding:
    :return:
    """
    if padding is None or padding == 'auto':
        padding = int(0.025*len(x))
        padding = max(padding, 1)
    if _np.isnan(x_hat).any():
        return _np.nan, _np.nan, _np.nan

    # RMS dxdt
    if dxdt_truth is not None:
        rms_dxdt = __rms_error__(dxdt_hat[padding:-padding], dxdt_truth[padding:-padding])
    else:
        rms_dxdt = None

    # RMS x
    if x_truth is not None:
        rms_x = __rms_error__(x_hat[padding:-padding], x_truth[padding:-padding])
    else:
        rms_x = None

    # RMS reconstructed x
    rec_x = _utility.integrate_dxdt_hat(dxdt_hat, dt)
    x0 = _utility.estimate_initial_condition(x, rec_x)
    rec_x = rec_x + x0
    rms_rec_x = __rms_error__(rec_x[padding:-padding], x[padding:-padding])

    return rms_rec_x, rms_x, rms_dxdt


def error_correlation(dxdt_hat, dxdt_truth, padding=None):
    """
    :param dxdt_hat:
    :param dxdt_truth:
    :param padding:
    :return:
    """
    if padding is None or padding == 'auto':
        padding = int(0.025*len(dxdt_hat))
        padding = max(padding, 1)
    errors = (dxdt_hat[padding:-padding] - dxdt_truth[padding:-padding])
    r = _scipy_stats.linregress(dxdt_truth[padding:-padding] -
                                _np.mean(dxdt_truth[padding:-padding]), errors)
    return r.rvalue**2


def rmse(dxdt_hat, dxdt_truth, padding=None):
    """
    :param dxdt_hat:
    :param dxdt_truth:
    :param padding:
    :return:
    """
    if padding is None or padding == 'auto':
        padding = int(0.025*len(dxdt_hat))
        padding = max(padding, 1)
    RMSE = _np.sqrt(_np.mean((dxdt_hat[padding:-padding] - dxdt_truth[padding:-padding])**2))
    return RMSE



"""
Unit tests for finite difference methods
"""
# pylint: skip-file
import numpy as np
from unittest import TestCase
from pynumdiff import first_order, second_order


class TestFD(TestCase):
    def test_first_order_1(self):
        x = np.array([1, 2, 3, 4, 5])
        dt = 0.01
        dxdt = np.array([100, 100, 100, 100, 100])
        _, dxdt_hat = first_order(x, dt)
        np.testing.assert_array_equal(dxdt_hat, dxdt)

    def test_first_order_2(self):
        x = np.array([8, 3, 14, 0, 9])
        dt = 0.01
        dxdt = np.array([-500, 300, -150, -250, 900])
        _, dxdt_hat = first_order(x, dt)
        np.testing.assert_array_equal(dxdt_hat, dxdt)

    def test_first_order_iterative(self):
        x = np.random.rand(100)
        dt = 0.01
        params = [100]
        _, dxdt_hat = first_order(x, dt, params, options={'iterate': True})
        assert x.shape == dxdt_hat.shape

    def test_second_order_1(self):
        x = np.array([1, 2, 3, 4, 5])
        dt = 0.01
        dxdt = np.array([100, 100, 100, 100, 100])
        _, dxdt_hat = second_order(x, dt)
        np.testing.assert_array_equal(dxdt_hat, dxdt)

    def test_second_order_2(self):
        x = np.array([8, 3, 14, 0, 9])
        dt = 0.01
        dxdt = np.array([-1300, 300, -150, -250, 2050])
        _, dxdt_hat = second_order(x, dt)
        np.testing.assert_array_equal(dxdt_hat, dxdt)



"""
Import useful functions from _finite_difference module
"""
from pynumdiff.finite_difference._finite_difference import first_order
from pynumdiff.finite_difference._finite_difference import second_order



"""
Simulation and Control of a cruise ?
"""

import numpy as np
from pynumdiff.utils import utility

parameters = {'omega_m': 420,
              'T_m': 190,
              'beta': 0.4,
              'Cr': 0.01,
              'Cd': 0.32,
              'A': 2.4,
              'g': 9.8,
              'm': 3000,
              'rho': 1.3,
              'v_r': 30,
              'k_p': 2,
              'k_i': 2
              }


def triangle(iterations, dt):
    """
    :param iterations:
    :param dt:
    :return:
    """
    t = np.arange(0, iterations*dt, dt)
    continuous_x = np.sin(0.02*t*np.sqrt(t))

    # find peaks and valleys
    peaks, valleys = utility.peakdet(continuous_x, 0.1)

    # organize peaks and valleys
    if len(peaks) > 0:
        reversal_idxs = peaks[:, 0].astype(int).tolist()
        reversal_vals = peaks[:, 1].tolist()
    else:
        reversal_idxs = []
        reversal_vals = []
    if len(valleys) > 0:
        reversal_idxs.extend(valleys[:, 0].astype(int).tolist())
        reversal_vals.extend(valleys[:, 1].tolist())

    reversal_idxs.extend([0, len(continuous_x)-1])
    reversal_vals.extend([0, continuous_x[-1]])

    idx = np.argsort(reversal_idxs)
    reversal_idxs = np.array(reversal_idxs)[idx]
    reversal_vals = np.array(reversal_vals)[idx]
    reversal_ts = t[reversal_idxs]

    x = np.interp(t, reversal_ts, reversal_vals)
    x = np.matrix(x)

    return x


def effective_wheel_radius(v=20):
    """
    :param v:
    :return:
    """
    return v


def torque(omega):
    """
    :param omega:
    :return:
    """
    omega_m = parameters['omega_m']
    t_m = parameters['T_m']
    beta = parameters['beta']
    return t_m*(1 - beta*(omega / omega_m - 1)**2)


# pylint: disable-msg=too-many-locals
def step_forward(state_vals, disturbances, desired_v, dt):
    """
    :param state_vals:
    :param disturbances:
    :param desired_v:
    :param dt:
    :return:
    """
    p = state_vals[0, -1]
    v = state_vals[1, -1]
    theta = disturbances[2, -1]
    m = parameters['m']
    g = parameters['g']
    Cr = parameters['Cr']
    rho = parameters['rho']
    Cd = parameters['Cd']
    A = parameters['A']
    v_r = desired_v[0, -1]
    alpha_n = effective_wheel_radius(v)
    z = np.sum(desired_v[0, :] - state_vals[1, :])*dt
    k_p = parameters['k_p']
    k_i = parameters['k_i']
    u = k_p*(v_r-v) + k_i*z

    # rolling friction
    Fr = m*g*Cr*np.sign(v)
    # aerodynamic drag
    Fa = 0.5*rho*Cd*A*np.abs(v)*v
    # forces due to gravity
    Fg = m*g*np.sin(theta)
    # driving force
    Fd = alpha_n*u*torque(alpha_n*v)
    vdot = 1/m*(Fd - (Fr + Fa + Fg))
    new_state = np.matrix([[p + dt*v], [v + vdot*dt], [theta]])
    return new_state, np.matrix(u)


# disturbance
def hills(iterations, dt, factor):
    """
    :param iterations:
    :param dt:
    :param factor:
    :return:
    """
    return triangle(iterations, dt)*0.3/factor


# desired velocity
def desired_velocity(n, factor):
    """
    :param n:
    :param factor:
    :return:
    """
    return np.matrix([2/factor]*n)


def run(timeseries_length=4, dt=0.01):
    """
    :param timeseries_length:
    :param dt:
    :return:
    """
    t = np.arange(0, timeseries_length, dt)
    iterations = len(t)

    # hills
    disturbances = np.matrix(np.zeros([3, iterations+1]))
    h = hills(iterations+1, dt, factor=0.5*timeseries_length/2)
    disturbances[2, :] = h[:, 0:disturbances.shape[1]]

    # controls
    controls = np.matrix([[0]])

    # initial condition
    state_vals = np.matrix([[0], [0], [0]])

    # desired vel
    v_r = desired_velocity(iterations, factor=0.5*iterations*dt/2)

    for i in range(1, iterations+1):
        new_state, u = step_forward(state_vals, disturbances[:, 0:i], v_r[:, 0:i], dt)
        state_vals = np.hstack((state_vals, new_state))
        controls = np.hstack((controls, u))

    return state_vals[0:2, 1:], disturbances[2, 1:], controls
